/* For copyright information, see olden_v1.0/COPYRIGHT */

/* Modified for sequential use and later automatic parallelization
 *
 * Stefan Haenssgen  19-dec-97
 *
 * 19-dec-97    SEE COMMENTS IN BITONIC.C SECTION!
 *              Merged all into one source code
 *		Removed DUMB defines
 *		ANSIfied forward declarations
 *		Removed procnames
 *		Made sort/merge procedures' return value void
 *		 (added reference parameters)
 *
 */

/* =============== NODE STRUCTURE =================== */

struct node { 
  int value;
  struct node *left;
  struct node *right;
  };

typedef struct node HANDLE;

#define NIL ((HANDLE *) 0)
#ifdef FUTURES
#define makenode(procid) ALLOC(procid,sizeof(struct node))
#else
#define makenode(procid) mymalloc(sizeof(struct node))
#endif
/* For copyright information, see olden_v1.0/COPYRIGHT */

/* ========== PROCEDURE TYPES/NUMS ================== */


void Bisort(int *ret_val, HANDLE *root, int spr_val, int dir);
void Bimerge(int *ret_val, HANDLE *root, int spr_val, int dir);
void RandTree(HANDLE **ret_val, int n,int seed,int node,int level);

void SwapValLeft(HANDLE *l,HANDLE *r,HANDLE *ll,HANDLE *rl, int lval, int rval);
void SwapValRight(HANDLE *l,HANDLE *r,HANDLE *lr,HANDLE *rr, int lval, int rval);


/* For copyright information, see olden_v1.0/COPYRIGHT */

/* Modified for sequential use and later automatic parallelization
 *
 * Stefan Haenssgen  19-dec-97
 *
 * 19-dec-97    Removed Olden-specific directives/includes:
 *		 CM sizes, NDim, ...
 */


int mylog(int num)
{
  int j=0,k=1;
  
  while(k<num) { k*=2; j++; }
  return j;
} 

extern int flag;
int dealwithargs(int argc, char *argv[])
{
  int size;

  if (argc > 2)
    flag = atoi(argv[2]);	/* Debugging flag */
  else
    flag = 0;

  if (argc > 1)
    size = atoi(argv[1]);
  else
    size = 16;
  return size;
  
}

void filestuff()
{
}
/* For copyright information, see olden_v1.0/COPYRIGHT */

/* Modified for sequential use and later automatic parallelization
 *
 * Stefan Haenssgen  19-dec-97
 *
 * 19-dec-97    Removed Olden-specific directives/includes:
 *		 future_cell_int, FUTURE if/elses
 *		Removed __NDim, __NumNodes, and CM Timers
 *		Annotated tree building as NOPARALLEL
 *		ANSI-fied procedures, re-formatted headers
 *		Set all to ONEONLY semantics
 *		Removed all RETEST/TEST calls
 *		Removed value references for ex-future_cells
 *		Removed dead code (SwapValue etc)
 *		Added comments
 *		Eliminated swap.c
 *
 */


/* =================== PROGRAM bitonic===================== */
/* UP - 0, DOWN - 1 */

#define CONST_m1 10000
#define CONST_b 31415821
#define RANGE 100
#define NULL 0

#define chatting printf

#define put(a) chatting("%d",a)
#define puts(a) chatting(a)
#define mymalloc malloc


int flag=0;		/* Debugging output if flag=1 */

#define LocalNewNode(h,v) \
  \
{ \
    h = (HANDLE *) mymalloc(sizeof(struct node)); \
      h->value = v; \
	h->left = NIL; \
	  h->right = NIL; \
	  };

#define NewNode(h,v,procid) LocalNewNode(h,v)

/* NOPARALLEL */
/* NOSTATS */
void InOrder(HANDLE *h)	/* In Order tree output */
{
HANDLE *l, *r;
if ((h != NIL))
  {
   l = h->left;
   r = h->right;
   InOrder(l);
   chatting("%d @ 0x%x\n",h->value,h);
   InOrder(r);
  }
}
int mult(int p, int q)
{
	int p1, p0, q1, q0;
	
	p1=p/CONST_m1; p0=p%CONST_m1;
	q1=q/CONST_m1; q0=q%CONST_m1;
	return (((p0*q1+p1*q0) % CONST_m1)*CONST_m1+p0*q0);
}

int skiprand(int seed, int n)
/* Generate the nth random # */
{
  for (; n; n--) seed=random(seed);
  return seed;
}

int random(int seed)
{
  int tmp;
  tmp = (mult(seed,CONST_b)+1);
  return tmp;
}


/* NOSTATS */

void RandTree(HANDLE **ret_val, int n,int seed,int node,int level)
{
  int next_val;
  HANDLE *f_left;
  HANDLE *f_right;
  HANDLE *h;

  if (n > 1)
    {
      int newnode;

      newnode = node;
      seed = random(seed);
      next_val=seed % RANGE;
      NewNode(h,next_val,node);
      
      RandTree(&f_left, (n/2),seed,newnode,level+1);
      RandTree(&f_right, (n/2),skiprand(seed,(n)+1),node,level+1); /* NOTHREAD */

      /* NEEDRESULTS */
      
      h->left = f_left;
      h->right = f_right;
    }
  else {
    h = NIL;
  }

  *ret_val = h;
} 

void SwapValLeft(HANDLE *l,HANDLE *r,HANDLE *ll,HANDLE *rl, int lval, int rval)
{
  r->value = lval;
  r->left = ll;
  l->left = rl;
  l->value = rval;
} 


void SwapValRight(HANDLE *l,HANDLE *r,HANDLE *lr,HANDLE *rr, int lval, int rval)
{  
  r->value = lval;
  r->right = lr;
  l->right = rr;
  l->value = rval;
  /*chatting("Swap Val Right l 0x%x,r 0x%x val: %d %d\n",l,r,lval,rval);*/
} 


void Bisort(int *ret_val, HANDLE *root, int spr_val, int dir)
{ HANDLE *l;
  HANDLE *r;
  int val;
  /*chatting("bisort %x\n", root);*/
  if ((root->left == NIL))
    { 
     if (((root->value > spr_val) ^ dir))
        {
	  val = spr_val;
	  spr_val = root->value;
	  root->value =val;
	}
    }
  else 
    {
      int ndir;
      int tmpval1;
      int tmpval2;
      int tmpval3;
      
      l = root->left;
      r = root->right;
      val = root->value;
      /*chatting("root 0x%x, l 0x%x, r 0x%x\n", root,l,r);*/
      
      Bisort(&tmpval1, l,val,dir);
      ndir = !dir;
      
      Bisort(&tmpval2, r,spr_val,ndir); /* NOTHREAD */

      /* NEEDRESULTS */
      
      root->value=tmpval1;
      spr_val=tmpval2;
      
      Bimerge(&tmpval3, root,spr_val,dir); /* NOTHREAD */
      spr_val=tmpval3;
    }
  /*chatting("exit bisort %x\n", root);*/

  /* NEEDRESULTS */

  *ret_val = spr_val;
} 


void Bimerge(int *ret_val, HANDLE *root, int spr_val, int dir)
{ int rightexchange;
  int elementexchange;
  int temp;
  HANDLE *pl,*pll,*plr;
  HANDLE *pr,*prl,*prr;
  HANDLE *rl;
  HANDLE *rr;
  int rv,lv;


  /*chatting("enter bimerge %x\n", root);*/
  rv = root->value;
  pl = root->left;
  pr = root->right;
  rightexchange = ((rv > spr_val) ^ dir);
  if (rightexchange)
    {
      root->value = spr_val;
      spr_val = rv;
    }
  
  while ((pl != NIL))
    {
      /*chatting("pl = 0x%x,pr = 0x%x\n",pl,pr);*/
      lv = pl->value;
      pll = pl->left;
      plr = pl->right;
      rv = pr->value;
      prl = pr->left;
      prr = pr->right;
      elementexchange = ((lv > rv) ^ dir);
      if (rightexchange)
        if (elementexchange)
          { 
            SwapValRight(pl,pr,plr,prr,lv,rv);
            pl = pll;
            pr = prl;
          }
        else 
          { pl = plr;
            pr = prr;
          }
      else 
        if (elementexchange)
          { 
            SwapValLeft(pl,pr,pll,prl,lv,rv);
            pl = plr;
            pr = prr;
          }
        else 
          { pl = pll;
            pr = prl;
          }
    }
  if ((root->left != NIL))
    { 
      int value;
      int tmpval1;	/* For allowing us to make the procedure type void */
      int tmpval2;
      
      rl = root->left;
      rr = root->right;
      value = root->value;
      
      Bimerge(&tmpval1, rl,value,dir);
      
      Bimerge(&tmpval2, rr,spr_val,dir); /* NOTHREAD */
      
      /* NEEDRESULTS */

      spr_val=tmpval2;
      root->value=tmpval1;
    }
  /*chatting("exit bimerge %x\n", root);*/

  *ret_val = spr_val;

} 


void main(int argc, char *argv[])
{ HANDLE *h;
  unsigned long time1, time2, time3, time4;
  int sval;
  int tmpval;
  int height;
  int counter;
  int n;
  int result = 1;
   
  
  n = dealwithargs(argc,argv);

  chatting("Bisort with %d size\n", n);
  RandTree(&h, n,12345768,0,0);
  sval = random(245867) % RANGE;
  if (flag) {
    InOrder(h);
    chatting("%d\n",sval);
   }
  chatting("**************************************\n");
  chatting("BEGINNING BITONIC SORT ALGORITHM HERE\n");
  chatting("**************************************\n");

  Bisort(&tmpval, h,sval,0);
  sval=tmpval;
  if (flag) {
    chatting("Sorted Tree:\n"); 
    InOrder(h);
    chatting("%d\n",sval);
   }
  
  Bisort(&tmpval, h,sval,1);
  sval=tmpval;
  if (flag) {
    chatting("Sorted Tree:\n"); 
    InOrder(h);
    chatting("%d\n",sval);
   }

  exit(0);
} 

