#!/bin/awk
#
# make_plot_from_messreihe.awk
#
# Generiert Plot-Befehle fuer Gnuplot und filtert Ausgabe.
# Berechnet auch Speedup bezogen auf "neverever" Eintrag
# in der Datei.
# Erzeugt Einzelpunkt-Plots fuer alle relevanten Parameter
# (Speedup, Threadzahl, CPU-Last, usec, ssec)
# Filtert auch fehlerhafte Zeilen (usec=ssec=0)
#
# Ausgabe: Zeilen mit "PLOT:" oder mit "DATA:" davor, direkt
#	   mit grep filterbar. PLOT-Zeilen sind Gnuplot-Befehle,
#	   DATA-Zeilen sind fuer die Eingabe fuer Gnuplot
#
# Beispiels-Eingabe:
#
# PROG eigenvalue.c CPUS 4 STRAT always 1 SIZE u 9000 RUN 1 THREADS 1 TIME 1833.0u 0.0s 15:39 195%
# PROG eigenvalue.c CPUS 4 STRAT always 1 SIZE u 9000 RUN 2 THREADS 1 TIME 1833.0u 0.0s 15:40 194%
# PROG eigenvalue.c CPUS 4 STRAT always 1 SIZE u 9000 RUN 3 THREADS 1 TIME 1832.0u 0.0s 15:39 195%
# PROG eigenvalue.c CPUS 4 STRAT first 1 SIZE u 9000 RUN 1 THREADS 4 TIME 1836.0u 0.0s 9:28 323%
#
#
# Stefan Haenssgen  11-nov-97
#
# 10-nov-97	Angepasst aus plot*.awk von alten plots-aus-stats
#		Kein Mittelwert o.ae. sondern alle Datenpunkte zeigen
#		Auch "defekte" Daten (0s Laufzeit) uebernehmen, weil
#		 Aborts & Co schon vom Messreihen-Script behandelt werden,
#		 sonst wuerden sie nicht im Plot auftauchen...
# 19-nov-97	Debug: h:m:s Umrechnung jetzt OK, oops...
# 11-feb-98	Durchgezogene Linie bei Speedup n und Auslastung n00%
#
#
BEGIN{
  n         = 0;				# Datenpunkte
  x         = 0;				# X-Achse (mehrere Dots auf y)
  seqtime   = 0;
  benchname = "";
  size      = "";
  aftersize = 10;				# Position nach size-Parameter
  cpus	    = 0;
  stratlabels["active"]    = "a";
  stratlabels["keep"]      = "k";
  stratlabels["depth"]     = "d";
  stratlabels["first"]     = "f";
  stratlabels["never"]     = " N ";
  stratlabels["neverever"] = " NE";
  stratlabels["always"]    = " A ";
  stratclasses["active"]    = "ACTIVE";
  stratclasses["keep"]      = "KEEP";
  stratclasses["depth"]     = "DEPTH";
  stratclasses["first"]     = "FIRST";
  stratclasses["never"]     = "N";
  stratclasses["neverever"] = "NE";
  stratclasses["always"]    = "AL";
  maxthreads = 0;
  maxusecs   = 0;
  maxssecs   = 0;
  maxsecs    = 0;
  maxperc    = 0;
}
{
 if (substr($1,1,1) != "#") {			# Kommentare ignorieren
  if (benchname == "") {			# Benchmark-Namen aus Programm!
    benchname = $2;    				# z.B. "eigenvalue.c"
    ip = index(benchname, ".c");
    if (ip>0) {
      benchname = substr(benchname, 1, ip-1);	# ".c" absaegen wenn vorhanden
    }
  }
  if (size == "") {				# Groesse (kann " " enthalten!)
    size = $9;
    i = 10;
    while ($i != "RUN") {			# Groessenparameter mit "-"
      size = sprintf("%s-%s", size, $i);	# statt " " zusammenbauen
      i++;
    }
    aftersize = i;				# Pos. nach Groesse merken
  }
  if (cpus == 0) {				# CPU-Zahl merken
    cpus = $4;
  }

  # Zeile einlesen und in relevante Teile zerlegen
  #
  n++;
  hasparam      = 0;
  stratbase[n]  = $6
  stratparam[n] = $7
  if (stratbase[n] ~ /keep|active|first|depth/ ) {
    strat[n] = sprintf("%s_%d", stratbase[n], stratparam[n]);
    hasparam = 1;			      	# Strategieparameter dazu
  } else {
    strat[n] = stratbase[n];			#  bzw. nicht bei Never & Co
  }
  stratlabel = stratlabels[stratbase[n]];
  if (n>1) {					# X-Achse fortschreiben
    if (strat[n] != strat[n-1]) {		#  wenn Strategie anders
      x++;
    }
    if (stratbase[n] != stratbase[n-1]) {	# Andere Strategieklasse
      label[x]="";				#  -> Trennlinie in Graph
      stratclass[x]=stratclasses[stratbase[n]];	#  (s.u. bei Ausgabe)
      x++;
    }
  } else {					# Anfang der Daten -> X starten
    stratclass[1]=stratclasses[stratbase[n]];
    label[1]="";
    x=2;
  }
  stratclass[x] = stratclasses[stratbase[n]];	# Fuer Unterteilung in Graph
  if (hasparam) {				# (Uberschreibt mehrfach,
###    label[x] = sprintf("%s%02d", stratlabel, $7);#  aber einfacher als extra
    label[x] = sprintf("%d", $7);
  } else {					#  in jedem else-Zweig)
    label[x] = stratlabel;
  }
  xs[n] = x;					# Daten-Position merken
  threads[n] = $(aftersize+3) + 0.0;		# Threads etc merken
  usecs[n]   = $(aftersize+5);			# Endungen (s,u,%) wegsaegen
  usecs[n]   = substr(usecs[n],1,length(usecs[n])-1) + 0.0;
  ssecs[n]   = $(aftersize+6);
  ssecs[n]   = substr(ssecs[n],1,length(ssecs[n])-1) + 0.0;
  perc[n]    = $(aftersize+8);
  perc[n]    = substr(perc[n],1,length(perc[n])-1) + 0.0;
#  if ((usecs[n] == 0) && (ssecs[n] == 0)) {	# Kaputte Daten wenn Zeit=0
#    n=n-1;
#    printf("# Error: %s\n",$0);
#  }
  s    = $(aftersize+7);
  nd   = 0;
  t[0] = 0; t[1] = 0; t[2] = 0;
  st   = 1;
  for (i=length(s); i>=1; i--) {                # Zeit [hh:]mm:ss parsen
    c = substr(s,i,1);
    if (c == ":") {
      nd++;
      st=1;
    } else {
      t[nd] = t[nd] + st * (c + 0);             # Zehnerstellen hochzaehlen
      st = st * 10;
    }
  }
  secs[n] = t[0]+t[1]*60+t[2]*60*60;
  if (secs[n] == 0) {				# Sekunden < 1 auf 0.5 runden
    if ( usecs[n] + ssecs[n] > 0) {		#  wenn Messung OK war (s/u=0
      secs[n] = 0.5;				#  falls Abbruch wegen Threads)
    }
  }
  if (threads[n]>maxthreads) maxthreads=threads[n];
  if (usecs[n]  >maxusecs  ) maxusecs  =usecs[n];
  if (ssecs[n]  >maxssecs  ) maxssecs  =ssecs[n];
  if (secs[n]   >maxsecs   ) maxsecs   =secs[n];
  if (perc[n]   >maxperc   ) maxperc   =perc[n];

  if (strat[n] == "neverever") {		# Sequentielle Laufzeit von
    seqtime = secs[n];				#  neverever-Strategie nehmen
  }
 } # Kommentare
}
END{
  # Plot-Befehle fuer GNUPLOT erzeugen
  #
print "PLOT:# maxperc", maxperc," maxusecs", maxusecs, "maxssecs",maxssecs, "maxsecs", maxsecs, "maxthreads", maxthreads;

  border=1.07					# Y-Randabstand (1.10 = 10%)
  printf("PLOT:# Measurements %s, Size %s, %d CPUs, Sequential time %ds\n", benchname, size, cpus, seqtime);
  printf("PLOT:#\n");
  xtics = "set xtics (";	  		# Beschriftung bzw. Trennlinien
  arrownum = 0;
  for (i=1; i<x; i++) {				#  sammeln/erzeugen
    if (label[i] == "") {
      arrownum++;				# Werden spaeter unten fuer
      arrowxs[arrownum] = i;			#  Trennlinien 0..max benutzt
    } else {
      xtics = sprintf("%s \"%s\" %d,", xtics, label[i], i);
    }
  }
  xtics = sprintf("%s \"%s\" %d)", xtics, label[x], x);
  printf("PLOT:%s\n", xtics);
  printf("PLOT:set tics out\n");
  printf("PLOT:set grid\n");
  printf("PLOT:set xrange [1:%d]\n", x+1);
  printf("PLOT:set term postscript eps 9\n");


  printf("PLOT:# Speedup:\n");
  printf("PLOT:set out \"%s_%s_%scpu_speedup.eps\"\n", benchname, size, cpus);
  max = (cpus+1)*border;
  printf("PLOT:set yrange [0:%f]\n", max);	# Speedup 0..CPUs+1
  for (i=1; i<=arrownum; i++) {
    printf("PLOT:set arrow %d from %d,0 to %d,%f nohead\n", i, arrowxs[i], arrowxs[i], max);
    printf("PLOT:set label %d \"%s\" at %f,%f\n", i, stratclass[arrowxs[i]], arrowxs[i]+n/1000, max/border*1.05);
  }
  printf("PLOT: set arrow %d from 0,%d to %d,%d nohead\n", arrownum+1, cpus, x+1, cpus);
  printf("PLOT:set title \"Speedup: %s (%s, %d CPUs)\"\n", benchname, size, cpus);
  printf("PLOT:set xlabel \"Strategy\"\n");
  printf("PLOT:set ylabel \"Speedup\"\n");
  printf("PLOT:set nokey\n");
  printf("PLOT:plot \"datafile\" using 1:8 title \"speedup\"\n", benchname, size, cpus);


  printf("PLOT:# Threads:\n");
  printf("PLOT:set out \"%s_%s_%scpu_threads.eps\"\n", benchname, size, cpus);
  max = maxthreads*border;
  printf("PLOT:set yrange [0:%f]\n", max);
  for (i=1; i<=arrownum; i++) {
    printf("PLOT:set arrow %d from %d,0 to %d,%f nohead\n", i, arrowxs[i], arrowxs[i], max);
    printf("PLOT:set label %d \"%s\" at %f,%f\n", i, stratclass[arrowxs[i]], arrowxs[i]+n/1000, max/border*1.05);
  }
  printf("PLOT:set ylabel \"Thread number\"\n");
  printf("PLOT:set title \"Number of Threads: %s (%s, %d CPUs)\"\n", benchname, size, cpus);
  printf("PLOT:plot \"datafile\" using 1:3\n");


  printf("PLOT:# utime, stime:\n");
  printf("PLOT:set out \"%s_%s_%scpu_ustime.eps\"\n", benchname, size, cpus);
  printf("PLOT:set title \"User/System Time: %s (%s, %d CPUs)\"\n", benchname, size, cpus);
  if (maxusecs > maxssecs) max = maxusecs*border
  else max = maxssecs*border;
  printf("PLOT:set yrange [0:%f]\n", max);
  for (i=1; i<=arrownum; i++) {
    printf("PLOT:set arrow %d from %d,0 to %d,%f nohead\n", i, arrowxs[i], arrowxs[i], max);
    printf("PLOT:set label %d \"%s\" at %f,%f\n", i, stratclass[arrowxs[i]], arrowxs[i]+n/1000, max/border*1.05);
  }
  printf("PLOT:set ylabel \"Time (s)\"\n");
  printf("PLOT:set key\n");
  printf("PLOT:plot \"datafile\" using 1:4 title \"utime\", \"datafile\" using 1:5 title \"stime\"\n");


  printf("PLOT:# Load:\n");
  printf("PLOT:set out \"%s_%s_%scpu_load.eps\"\n", benchname, size, cpus);
  printf("PLOT:set title \"System Load: %s (%s, %d CPUs)\"\n", benchname, size, cpus);
  # max = maxperc*border;
  max = (cpus+0.3)*100*border;		# Last 0..CPU*100%
  printf("PLOT:set yrange [0:%f]\n", max);
  for (i=1; i<=arrownum; i++) {
    printf("PLOT:set arrow %d from %d,0 to %d,%f nohead\n", i, arrowxs[i], arrowxs[i], max);
    printf("PLOT:set label %d \"%s\" at %f,%f\n", i, stratclass[arrowxs[i]], arrowxs[i]+n/1000, max/border*1.05);
  }
  printf("PLOT: set arrow %d from 0,%d to %d,%d nohead\n", arrownum+1, cpus*100, x+1, cpus*100);
  printf("PLOT:set ylabel \"Load (%%)\"\n");
  printf("PLOT:set nokey\n");
  printf("PLOT:plot \"datafile\" using 1:7 title \"Load %%\"\n");
  printf("PLOT:\n");


  printf("DATA:# Measurements %s, Size %s, %d CPUs, Sequential time %ds\n", benchname, size, cpus, seqtime);
  printf("DATA:# Format: 1=X     2=strategy 3=threads 4=u_sec\n");
  printf("DATA:#         5=s_sec 6=wall_sec 7=load%%   8=speedup\n");
  for (i=1; i<=n; i++) {
    if (secs[i]>0) {
      speedup = seqtime/secs[i];
    } else {
      speedup = 0;   				# secs=0 bei Abort-Run
    }
    printf("DATA: %2d %-9s %5d %9.2f %9.2f %8.1f %3d %6.3f\n", xs[i], strat[i], threads[i], usecs[i], ssecs[i], secs[i], perc[i], speedup);
  }
}

